local M = {}

local scenario = require("scenario/scenarios")
local helper = require("scenario/scenariohelper")

local playerName = "rl_player0"
local trafficNames = {"rl_traffic0", "rl_traffic1", "rl_traffic2"}
local failMessages = {busted = "Busted! Get in the handcuffs.", crashBusted = "Totaled... and busted!", crash = "Totaled!"}
local winMessages = {normal = "Phew, you did it! Enjoy the festival!", lawful = "Phew, you did it! Plus, you're not wanted by the cops! Nice driving, champ.", damaged = "Phew, you did it! Wow, your vehicle got really beat up. Maybe you should also sign up for the demolition derby!", veryFast = "Phew, you did it! What a blazing fast time! Your heart must be racing.", ultraFast = "Phew, you did it! Holy fudge, what an incredibly insane time! I bet you intimidated the local street racers with that run. Fantastic driving!", almost = "Phew, you did it! Just in time, they were starting to close the doors!", policeCar = "Phew, you did it! In a police vehicle, I see. Smart choice!", funnyCar = "Phew, you did it! Wait, what? With that vehicle? Wow, you're insane, but you've earned a headline in the newspaper! I'm impressed.", evilCar = "Phew, you did it! You're kinda evil, though. A true millionaire villain."}

local zoneNames = {}
local player = {}
local traffic = {}
local Traffic = {}

local queue = 0
local speedLimit = 0
local mainTimer = 0

local trafficTotal = 3 --Number of traffic cars in the prefab
local zonesTotal = 28 --Number of spawn zoneNames in the prefab
local minSpeed = 16 --Minimum speed limit for randomizer, in metres per second
local maxSpeed = 20 --Maximum speed limit for randomizer, in metres per second
local speedingSpeed = 28 --Excessive speed!
local spawnWay = 25 --Percent chance that the traffic car will spawn going forwards
local stopForwardSpawn = 25 --Zone where traffic won't spawn going forwards anymore
local trafficAggro = 0.85 --The aggression level of the traffic cars
local playerSmash = 45000 --The amount of damage on the player car that will fail the scenario
local trafficSmash = 2000 --The amount of damage on a traffic car that will enable it to reinitialize
local nearDist = 20 --The distance of a traffic car that is considered overtaken by the player
local farDist = 300 --The distance of a traffic car that is too far away to have any fun with the player
local readyDist = 120 --The distance of a traffic car that is ready to test for teleportation, & the distance that cops can see you
local farMultiplier = 2.5 --Spawn test multipler for distant traffic cars; used with readyDist
local policeCar = 3 --Which AI vehicle is the police car?
local policeSpawn = 35 --Police car will be ready after this many seconds
local randomColors = false --If true, traffic car colors can get changed to whatever!

function Traffic:new(obj, color, pos, damage, state, role, travel, zone, far, timer)
	local self = setmetatable({}, Traffic)
	self.__index = self
	self.obj = obj or nil --The vehicle itself
	self.color = color or nil --Array of colors to pick from for repainting the car; nests into RGBA arrays
	self.pos = pos or nil --Vehicle position
	self.damage = damage or 0 --Vehicle damage points
	self.state = state or nil --Teleport states; none, normal, ready, moved
	self.role = role or nil --Purpose of the traffic car; normal, policeNormal, policeChase
	self.travel = travel or nil --Direction of travel; none, backward, forward
	self.zone = zone or nil --Zone that the vehicle is currently in
	self.far = far or 1 --Multiplier for spawn mode; see below
	self.timer = timer or 0 --Internal logic timer
	return self
end

local function setSpawn(aTraffic) --Sets the new spawn point of a traffic car and rotates it accordingly
	local allow = true
	local randomWay = math.random(1, 100)
	local zoneIn = scenetree.findObject(zoneNames[player.zone + 3]) --This spawn zone is actually three away from the last spawn zone that the player passed; minimizes sudden appearing
	local zonePos = zoneIn:getPosition()
	local zoneRot = zoneIn:getRotation()
	local colorValues = {}
	local colorIn = aTraffic.color[math.random(1, #aTraffic.color)]
	if aTraffic.role == "policeNormal" then
		randomWay = math.random(1, math.ceil(spawnWay * 1.2))
	end
	if player.zone + 3 >= stopForwardSpawn then
		randomWay = 100
	end
	for i, v in ipairs(traffic) do
		if traffic[i].zone == player.zone + 3 then
			allow = false	
		end
	end
	if randomWay <= spawnWay then
		zoneRot = quatFromDir(vec3(0,1,0):rotated(zoneRot):rotated(quat(0, 0, -1, 0)), vec3(0,0,1)) --Yay, it works! This rotates the spawned traffic car by 180 degrees
	end
	if allow == true then
		if randomWay <= spawnWay then
			aTraffic.travel = "forward"
		else
			aTraffic.travel = "backward"
		end
		aTraffic.obj:setPositionRotation(zonePos.x, zonePos.y, zonePos.z + 0.5, 0, 0, zoneRot.z, zoneRot.w + (randomWay / 1000)) --The random addition fixes a spawn rotation init bug
		log("Z", "gull", aTraffic.obj.name.." teleported to "..zoneIn.name)
		if aTraffic.damage >= trafficSmash then
			aTraffic.obj:queueLuaCommand("vehicleResetted(false)")
		end
		if aTraffic.role == "policeChase" then
			aTraffic.obj:queueLuaCommand("electrics.set_lightbar_signal(2)")
			player.policeWatch = "chase"
		end
		if randomColors == true then
			colorValues = {math.random(), math.random(), math.random(), math.random() + 1}
		else
			colorValues = {colorIn[1], colorIn[2], colorIn[3], colorIn[4]}
		end
		aTraffic.obj.color = ColorF(colorValues[1], colorValues[2], colorValues[3], colorValues[4]):asLinear4F()
		aTraffic.obj.colorPalette0 = ColorF(colorValues[1], colorValues[2], colorValues[3], colorValues[4]):asLinear4F()
		aTraffic.state = "moved"
		aTraffic.timer = 0
	end
end

local function arrestPlayer() --Arrests the player with a countdown
	if player.arrestTimer % 1 == 0 then
		local countdown = 3 - math.modf(player.arrestTimer)
		if countdown > 0 then
			helper.flashUiMessage((tostring(countdown).."..."), 0.5, true)
		end
	end
	player.arrestTimer = player.arrestTimer + 0.25
	if player.arrestTimer > 3 then
		player.success = false
		scenario.finish({failed = failMessages.busted})
	end
end

local function chasePlayer(aPolice) --Various logic for while the police car is chasing the player
	local avoid = false
	local chaseDist = readyDist / 2
	local playerDist = (player.pos - aPolice.pos):len()
	aPolice.far = 2
	for i, v in ipairs(traffic) do
		if traffic[i].role == "normal" then
			traffic[i].pos = traffic[i].obj:getPosition()
			local policeDist = (aPolice.pos - traffic[i].pos):len()
			if policeDist <= chaseDist and playerDist <= readyDist then --Traffic reacts to active AI police
				traffic[i].obj:queueLuaCommand("ai.setSpeed(5)")
			end
			if aPolice.travel == "forward" and traffic[i].travel == "forward" and policeDist <= chaseDist then --AI police will try to dodge traffic
				avoid = true
			end
		end
	end
	if playerDist <= chaseDist then
		player.arrestWatch = true
		if (aPolice.travel == "backward" or playerDist <= nearDist) and player.policeWatch == "chase" then
			aPolice.obj:queueLuaCommand("ai.setMode('chase')")
			aPolice.obj:queueLuaCommand("ai.driveInLane('off')")
			player.policeWatch = "force"
		end
		if player.vel <= 4 then
			arrestPlayer()
		else
			player.arrestTimer = 0
		end
	else
		player.arrestWatch = false
		if aPolice.travel == "backward" and player.policeWatch == "force" then
			aPolice.obj:queueLuaCommand("ai.setMode('random')")
			aPolice.obj:queueLuaCommand("ai.driveInLane('on')")
			player.policeWatch = "chase"
		end
	end
	if avoid == false then
		if player.policeWatch == "chase" then
			aPolice.obj:queueLuaCommand("ai.setSpeed("..tostring(math.max(player.vel * ((readyDist / math.min(readyDist, playerDist) / 10) + 0.6), 5))..")")
		end
		if player.policeWatch == "force" then
			aPolice.obj:queueLuaCommand("ai.setSpeed("..tostring(math.max(player.vel * 1.5, 5))..")")
		end
	else
		aPolice.obj:queueLuaCommand("ai.setSpeed("..tostring(speedingSpeed / 2)..")")
		aPolice.obj:queueLuaCommand("ai.driveInLane('off')")
	end
end

local function onScenarioUIReady()
	player = {obj = nil, pos = nil, vel = nil, dist = nil, damage = nil, damageFail = nil, zone = nil, policeWatch = nil, arrestWatch = nil, arrestTimer = nil, number = nil, crazy = nil, success = nil}
	player.obj = scenetree.findObject(playerName)
	player.damageFail = playerSmash
	player.zone = 0
	player.policeWatch = "none"
	player.arrestWatch = false
	player.arrestTimer = 0
	player.number = 0
	player.crazy = 0
	player.success = true
	queue = 0
	mainTimer = 0
	for  i = 1, zonesTotal, 1 do
		zoneNames[i] = "spawnzone"..(i - 1)
	end
	for i, v in ipairs(trafficNames) do
		traffic[i] = Traffic:new(scenetree.findObject(trafficNames[i]), {}, nil, 0, "normal", "normal", "none", "none", farMultiplier, 0) --obj, color, pos, damage, state, role, travel, zone, far, timer
		local zonePos = scenetree.findObject(zoneNames[i]):getPosition()
		local zoneRot = scenetree.findObject(zoneNames[i]):getRotation()
		traffic[i].obj:setPositionRotation(zonePos.x, zonePos.y, zonePos.z, 0, 0, zoneRot.z, zoneRot.w)
		traffic[i].travel = "backward"
		if i == policeCar then --AI police car
			zonePos = scenetree.findObject(zoneNames[math.ceil(zonesTotal / 1.5)]):getPosition()
			zoneRot = scenetree.findObject(zoneNames[math.ceil(zonesTotal / 1.5)]):getRotation()
			traffic[i].obj:setPositionRotation(zonePos.x, zonePos.y, zonePos.z, 0, 0, zoneRot.z, zoneRot.w)
			traffic[i].state = "none"
			traffic[i].role = "policeNormal"
		end
	end
	traffic[1].color[1] = {0.28, 0.04, 0.02, 1.5}
	traffic[1].color[2] = {0.08, 0.08, 0.08, 1.5}
	traffic[1].color[3] = {0.32, 0.04, 0.02, 1.5}
	traffic[1].color[4] = {0.7, 0.7, 0.7, 1.2}
	
	traffic[2].color[1] = {0.02, 0.02, 0.02, 1.4}
	traffic[2].color[2] = {0.25, 0.04, 0.03, 1.4}
	traffic[2].color[3] = {0.54, 0.49, 0.32, 1.4}
	traffic[2].color[4] = {0.02, 0.03, 0.15, 1.4}
	
	traffic[3].color[1] = {0.96, 0.96, 0.96, 1.6}
end

local function onRaceStart()
	local policeVehicles = {"vehicles/coupe/police.pc", "vehicles/super/polizia.pc", "vehicles/etk800/police.pc", "vehicles/fullsize/police.pc", "vehicles/fullsize/bcpd.pc", "vehicles/etkc/police_highway_A.pc", "vehicles/autobello/carabinieri.pc", "vehicles/burnside/police.pc", "vehicles/sunburst/police.pc", "vehicles/sunburst/police_RS.pc", "vehicles/van/police.pc", "vehicles/roamer/sheriff.pc", "vehicles/roamer/sheriff_offroad.pc"}
	local crazyVehicles = {"ball", "barrels", "barrier", "blockwall", "bollard", "boxutility", "boxutility_large", "cannon", "caravan", "christmas_tree", "citybus", "cones", "dryvan", "flail", "flatbed", "flipramp", "haybale", "inflated_mat", "kickplate", "large_angletester", "large_bridge", "large_cannon", "large_crusher", "large_hamster_wheel", "large_metal_ramp", "large_roller", "large_spinner", "large_tilt", "metal_box", "metal_ramp", "piano", "pigeon", "roadsigns", "rocks", "rollover", "sawhorse", "semi", "streetlight", "suspensionbridge", "tanker", "tirestacks", "tirewall", "trafficbarrel", "tsfb", "tube", "van", "wall", "weightpad", "woodcrate", "woodplanks"}
	local ramplow = "vehicles/semi/t65_ramplow.pc"
	player.number = 1
	for i, v in ipairs(policeVehicles) do
		if player.obj.partConfig == policeVehicles[i] then
			player.crazy = 1
			player.number = 1.4 --Modifies speed limit, police start spawn time, police speed watch
		end
	end
	for i, v in ipairs(crazyVehicles) do
		if player.obj.jbeam == crazyVehicles[i] and player.crazy ~= 1 then
			player.crazy = 2
			player.number = 1.1
			if player.obj.jbeam ~= "pigeon" then
				player.damageFail = 100000 --Damage workaround for heavy vehicles
			end
		end
	end
	if player.obj.partConfig == ramplow then
		player.crazy = 3
		player.number = 0.6
	end
	speedLimit = math.ceil(math.random(minSpeed, maxSpeed) / player.number) --Creates a new speed limit for all traffic cars; makes restarts more interesting but prevents traffic cars from bumping into each other
	log("Z", "gull", "Speed limit: "..speedLimit)
	for i, v in ipairs(traffic) do
		traffic[i].obj:queueLuaCommand("ai.setAggression("..tostring(trafficAggro)..")")
		traffic[i].obj:queueLuaCommand("ai.driveInLane('on')")
		traffic[i].obj:queueLuaCommand("ai.setSpeed("..tostring(speedLimit)..")")
		traffic[i].obj:queueLuaCommand("ai.setSpeedMode('limit')")
		traffic[i].obj:queueLuaCommand("ai.resetLearning(true)")
		traffic[i].obj:queueLuaCommand("ai.setMode('random')")
	end
end

local function onRaceTick(raceTickTime)
	mainTimer = mainTimer + raceTickTime
	player.pos = player.obj:getPosition()
	player.vel = map.objects[map.objectNames[player.obj.name]].vel:length()
	player.damage = map.objects[map.objectNames[player.obj.name]].damage
	for i, v in ipairs(traffic) do
		traffic[i].pos = traffic[i].obj:getPosition()
		player.dist = (player.pos - traffic[i].pos):len()
		traffic[i].damage = map.objects[map.objectNames[traffic[i].obj.name]].damage
		if traffic[i].state == "moved" then
			if traffic[i].timer >= 0.5 then --Any traffic timers, such as here, intentionally delay AI mode changes; experimental!
				traffic[i].obj:queueLuaCommand("ai.setMode('random')")
				traffic[i].obj:queueLuaCommand("ai.driveInLane('on')")
				traffic[i].state = "normal"
				traffic[i].far = 1
			end
			traffic[i].timer = traffic[i].timer + raceTickTime
		end
		if mainTimer >= policeSpawn * player.number then
			if traffic[i].role == "policeNormal" and player.policeWatch == "none" then
				traffic[i].state = "normal"
				player.policeWatch = "patrol"
			end
		end
		if player.dist <= nearDist then --If you pass a traffic car, prepare it for teleportation
			traffic[i].state = "ready"
			traffic[i].far = 1
			traffic[i].timer = 0
		end
		if player.dist <= readyDist / 2 then
			if traffic[i].role == "policeNormal" and player.policeWatch == "patrol" and (player.vel >= speedingSpeed * player.number or (player.dist <= 8 and traffic[i].damage >= 1000)) then
				traffic[i].obj:queueLuaCommand("electrics.set_lightbar_signal(2)")
				traffic[i].obj:queueLuaCommand("ai.setAggression(1.2)")
				traffic[i].role = "policeChase"
				player.policeWatch = "chase"
			end
		end
		if player.zone > 1 then --Enables traffic spawning and other behaviors after you pass enough zones
			if player.dist >= farDist and traffic[i].state == "normal" then --If you get really far away from a traffic car, prepare it for teleportation
				traffic[i].state = "ready"
				traffic[i].far = farMultiplier --Multiplier for ready distance; can prevent sudden vanishing
				traffic[i].timer = 0
			end
			if queue > 0 then
				if player.dist >= (readyDist * traffic[i].far) and traffic[i].state == "ready" then
					if traffic[i].timer == 0 and traffic[i].zone == "none" then
						traffic[i].obj:queueLuaCommand("ai.setMode('disabled')")
						traffic[i].far = 0
					end
					if player.dist >= readyDist and traffic[i].timer >= 0.5 and traffic[i].obj.name == traffic[queue].obj.name then
						setSpawn(traffic[i])
						queue = -4 --Artificial delay for next spawn test; prevents lag and stuff
					end
					traffic[i].timer = traffic[i].timer + raceTickTime
				end
			end
		end
		if traffic[i].role == "policeChase" then
			chasePlayer(traffic[i])
		end
	end
	queue = queue + 1 --This gets updated every race tick to prevent two cars from being teleported at the same time to the same spawn zone
	if queue > trafficTotal then
		queue = 1
	end
	if mainTimer == 3 then
		if player.crazy >= 2 then
			helper.flashUiMessage("You picked this vehicle? Are you insane?", 3, false)
		end
	end
	if player.damage >= player.damageFail then
		player.success = false
		if player.arrestWatch == true then
			scenario.finish({failed = failMessages.crashBusted})
		else
			scenario.finish({failed = failMessages.crash})
		end	
	end
	if mainTimer >= 140 then
		player.success = false
	end
end

local function onRaceResult()
	if player.success == true then
		if player.crazy == 1 then --Chaotic good
			scenario.finish({msg = winMessages.policeCar})
		elseif player.crazy == 2 then --Chaotic neutral
			scenario.finish({msg = winMessages.funnyCar})
		elseif player.crazy == 3 then --Chaotic evil
			scenario.finish({msg = winMessages.evilCar})
		elseif mainTimer <= 90 then --Lawful evil
			scenario.finish({msg = winMessages.ultraFast})
		elseif mainTimer <= 100 then --Lawful neutral
			scenario.finish({msg = winMessages.veryFast})
		elseif mainTimer >= 135 then --Neutral good
			scenario.finish({msg = winMessages.almost})
		elseif player.damage >= player.damageFail / 2 then --Neutral evil
			scenario.finish({msg = winMessages.damaged})
		elseif player.policeWatch == "patrol" then --Lawful good
			scenario.finish({msg = winMessages.lawful})
		else --True neutral
			scenario.finish({msg = winMessages.normal})
		end
	end
end

local function onBeamNGTrigger(data)
	if data.event == "enter" then
		if data.subjectName == player.obj.name and string.byte(tostring(data.triggerName)) == 115 then --Prevents stupid trigger confusion with bus stops
			for i, v in ipairs(zoneNames) do
				if data.triggerName == zoneNames[i] then
					player.zone = i
				end
			end
			if player.zone + 3 > zonesTotal then --Prevents spawn zone errors
				player.zone = 0
			end
			log("Z", "gull", "Zone check: "..player.zone)
		end
		for i, v in ipairs(traffic) do
			if data.subjectName == traffic[i].obj.name then
				if string.byte(tostring(data.triggerName)) == 105 then --intersections
					traffic[i].obj:queueLuaCommand("ai.setSpeed(8)")
				end
				for o, v in ipairs(zoneNames) do
					if data.triggerName == zoneNames[o] then
						traffic[i].zone = o
					end
				end	
				if data.triggerName == zoneNames[player.zone + 3] then
					if traffic[i].role ~= "policeChase" then
						traffic[i].obj:queueLuaCommand("ai.setSpeed(10)") --Traffic cars slow down in zones where teleports and stuff are likely
						log("Z", "gull", traffic[i].obj.name.." is aware in zone "..data.triggerName)
					end
				end
				log("Z", "gull", traffic[i].obj.name.." entered zone "..data.triggerName)
			end
		end
	end
	if data.event == "exit" then
		for i, v in ipairs(traffic) do
			if data.subjectName == traffic[i].obj.name then
				if string.byte(tostring(data.triggerName)) == 115 then
					if traffic[i].state ~= "moved" then
						traffic[i].zone = "none"
					end
					if traffic[i].role ~= "policeChase" then
						traffic[i].obj:queueLuaCommand("ai.setSpeed("..tostring(speedLimit)..")")
					end
				end
				log("Z", "gull", traffic[i].obj.name.." exited zone "..data.triggerName)
			end
		end
	end
end

M.onScenarioUIReady = onScenarioUIReady
M.onRaceStart = onRaceStart
M.onRaceTick = onRaceTick
M.onRaceResult = onRaceResult
M.onBeamNGTrigger = onBeamNGTrigger

return M